/*-
 * See the file LICENSE for redistribution information.
 *
 * Copyright (c) 2002-2009 Oracle.  All rights reserved.
 *
 * $Id: RollbackException.java,v 1.7 2009/05/11 12:36:05 cwl Exp $
 */

package com.sleepycat.je.rep;

import static com.sleepycat.je.dbi.EnvironmentFailureReason.HARD_RECOVERY;

import com.sleepycat.je.Database;
import com.sleepycat.je.rep.impl.RepImpl;
import com.sleepycat.je.utilint.DbLsn;
import com.sleepycat.je.utilint.VLSN;

/**
 * This asynchronous exception indicates that a new master has been selected,
 * this <code>Replica</code>'s log is ahead of the current <code>Master</code>,
 * and in this case, the <code>Replica</code> was unable to rollback without a
 * recovery. As a consequence, one or more of the most recently committed
 * transactions may need to be rolled back, before the <code>Replica</code> can
 * synchronize its state with that of the current <code>Master</code>.
 * <p>
 * Existing {@link ReplicatedEnvironment}, and consequently {@link Database}
 * handles, are invalidated as a result of this exception. The application must
 * close all old handles and create new handles before it can proceed. The
 * actual rollback of any the recently committed transactions is done when the
 * application reinstantiates and thereby reopens the {@link
 * ReplicatedEnvironment}.  The application is responsible for discarding and
 * recreating any transient state that may be associated with the committed
 * transactions that were rolled back.
 * <p>
 * This exception should be encountered relatively infrequently in practice,
 * since the election mechanism favors nodes with the most advanced log when
 * deciding upon a master. The exception, due to its nature, can only be
 * encountered when the node is in the <code>Replica</code> state, or the node
 * is trying to transition to the <code>Replica</code> state.
 * <p>
 * Use of weak durability requirements like
 * {@link com.sleepycat.je.Durability.ReplicaAckPolicy#NONE} increases the
 * likelihood of this exception.
 */
@SuppressWarnings("serial")
public class RollbackException extends RestartRequiredException {

    /* 
     * TODO: it would be nicer to provide a set of all the discarded txn ids.
     * But is that really useful for the application, or too much data?
     */
    private final VLSN lastTxnEnd;
    private final VLSN matchpointVLSN;
    private final long matchpointLsn;
    private final String nodeName;

    /**
     * For internal use only.
     * @hidden
     */
    public RollbackException(RepImpl repImpl,
                             VLSN lastTxnEnd,
                             VLSN matchpointVLSN,
                             long matchpointLsn) {
        super(repImpl, HARD_RECOVERY);
        this.lastTxnEnd = lastTxnEnd;
        this.matchpointVLSN = matchpointVLSN;
        this.matchpointLsn = matchpointLsn;
        this.nodeName = repImpl.getName();
    }

    @Override
    public String getMessage() {
        return "Node " + nodeName + " must be rolled back, recovery is " +
            "required. All existing ReplicatedEnvironment handles must be " +
            " closed and reinstantiated. Log files were truncated at file 0x" + 
            DbLsn.getFileNumber(matchpointLsn) + ", offset 0x" +
            DbLsn.getFileOffset(matchpointLsn) + ". LastTxnEnd=" + 
            lastTxnEnd + " matchpoint at " +  matchpointVLSN; 
    }

    /**
     * For internal use only.
     * @hidden
     */
    public RollbackException(String message, RollbackException cause) {
        super(message, cause);
        this.lastTxnEnd = cause.lastTxnEnd;
        this.matchpointVLSN = cause.matchpointVLSN;
        this.matchpointLsn = cause.matchpointLsn;
        this.nodeName = cause.nodeName;
    }

    /**
     * For internal use only.
     * @hidden
     */
    @Override
    public RollbackException wrapSelf(String msg) {
        return new RollbackException(msg, this);
    }
}
