/*-
 * See the file LICENSE for redistribution information.
 *
 * Copyright (c) 2002-2009 Oracle.  All rights reserved.
 *
 * $Id: MonitorService.java,v 1.14 2009/09/15 23:28:26 linda Exp $
 */
package com.sleepycat.je.rep.monitor;

import java.io.IOException;
import java.nio.channels.SocketChannel;
import java.util.logging.Formatter;
import java.util.logging.Level;
import java.util.logging.Logger;

import com.sleepycat.je.rep.ReplicationGroup;
import com.sleepycat.je.rep.impl.TextProtocol.RequestMessage;
import com.sleepycat.je.rep.impl.TextProtocol.ResponseMessage;
import com.sleepycat.je.rep.monitor.Protocol.GroupChange;
import com.sleepycat.je.rep.utilint.ReplicationFormatter;
import com.sleepycat.je.rep.utilint.ServiceDispatcher;
import com.sleepycat.je.rep.utilint.ServiceDispatcher.ExecutingService;
import com.sleepycat.je.utilint.LoggerUtils;

/**
 * @hidden
 * For internal use only.
 */
public class MonitorService extends ExecutingService {
    private final Monitor monitor;
    private final Protocol protocol;
    private final Logger logger;
    private final Formatter formatter;

    /* Identifies the Group Service. */
    public static final String SERVICE_NAME = "Monitor";

    public MonitorService(Monitor monitor,
                          ServiceDispatcher dispatcher) {
        super(SERVICE_NAME, dispatcher);
        this.monitor = monitor;
        protocol = new Protocol(monitor.getGroupName(),
                                monitor.getMonitorNameIdPair(),
                                null);
        logger = LoggerUtils.getLoggerFormatterNeeded(getClass());
        formatter = new ReplicationFormatter(monitor.getMonitorNameIdPair());
    }

    /* Dynamically invoked process methods */

    /**
     * Notify the monitor about the new group
     */
    public ResponseMessage process(GroupChange groupChange) {
        monitor.notify(new ReplicationGroup(groupChange.getGroup()));
        return null;
    }

    @Override
    public Runnable getRunnable(SocketChannel socketChannel) {
        return new MonitorServiceRunnable(socketChannel);
    }

    class MonitorServiceRunnable implements Runnable {
        final SocketChannel channel;

        MonitorServiceRunnable(SocketChannel socketChannel) {
            this.channel = socketChannel;
        }

        public void run() {
            try {
                channel.configureBlocking(true);
                RequestMessage requestMessage =
                    protocol.getRequestMessage(channel);
                if (requestMessage == null) {
                    return;
                }
                ResponseMessage responseMessage =
                    protocol.process(MonitorService.this, requestMessage);
                assert(responseMessage == null);
            } catch (IOException e) {
                LoggerUtils.logMsg(logger, formatter, Level.WARNING, 
                                   "IO error on socket: " + e.getMessage());
                return;
            } finally {
                if (channel.isOpen()) {
                    try {
                        channel.close();
                    } catch (IOException e) {
                        LoggerUtils.logMsg
                            (logger, formatter, Level.WARNING,
                             "IO error on channel close: " + e.getMessage());
                        return;
                    }
                }
            }
        }
    }
}
