/*-
 * See the file LICENSE for redistribution information.
 *
 * Copyright (c) 2002-2009 Oracle.  All rights reserved.
 *
 * $Id: ReplicaFeederSyncup.java,v 1.45 2009/09/18 01:15:08 linda Exp $
 */
package com.sleepycat.je.rep.stream;

import static com.sleepycat.je.utilint.VLSN.NULL_VLSN;

import java.io.IOException;
import java.util.Arrays;
import java.util.HashSet;
import java.util.logging.Level;
import java.util.logging.Logger;

import com.sleepycat.je.DatabaseException;
import com.sleepycat.je.EnvironmentFailureException;
import com.sleepycat.je.config.EnvironmentParams;
import com.sleepycat.je.dbi.EnvironmentImpl;
import com.sleepycat.je.rep.InsufficientLogException;
import com.sleepycat.je.rep.ReplicationNode;
import com.sleepycat.je.rep.RollbackException;
import com.sleepycat.je.rep.impl.RepImpl;
import com.sleepycat.je.rep.impl.node.LocalCBVLSNTracker;
import com.sleepycat.je.rep.impl.node.RepNode;
import com.sleepycat.je.rep.impl.node.Replay;
import com.sleepycat.je.rep.stream.Protocol.AlternateMatchpoint;
import com.sleepycat.je.rep.stream.Protocol.Entry;
import com.sleepycat.je.rep.stream.Protocol.EntryNotFound;
import com.sleepycat.je.rep.stream.Protocol.RestoreResponse;
import com.sleepycat.je.rep.utilint.NamedChannel;
import com.sleepycat.je.rep.utilint.BinaryProtocol.Message;
import com.sleepycat.je.rep.vlsn.VLSNIndex;
import com.sleepycat.je.rep.vlsn.VLSNRange;
import com.sleepycat.je.utilint.DbLsn;
import com.sleepycat.je.utilint.LoggerUtils;
import com.sleepycat.je.utilint.VLSN;

/**
 * Establish where the replication stream should start for a replica and feeder
 * pair. The replica compares what is in its log with what the feeder has, to
 * determine the latest common log entry matchpoint
 *
 * - If the replica has applied log entries after that matchpoint, roll them
 *   back
 * - If a common matchpoint can't be found, the replica will need to do
 *   a network restore.
 */
public class ReplicaFeederSyncup {

    private final Logger logger;

    private final NamedChannel namedChannel;
    private final Protocol protocol;
    private final RepNode repNode;
    private final VLSNIndex vlsnIndex;
    private final Replay replay;
    private final RepImpl repImpl;

    /* The VLSN, lsn and log entry at which a match was made. */
    private VLSN matchpointVLSN = NULL_VLSN;
    private long matchpointLsn = DbLsn.makeLsn(0, 0);
    private Long matchedVLSNTime = 0L;
    private boolean passedCheckpointEnd = false;

    private static TestHook<Object> syncupEndHook;

    public ReplicaFeederSyncup(RepNode repNode,
                               Replay replay,
                               NamedChannel namedChannel,
                               Protocol protocol) {
        this.replay = replay;
        logger = LoggerUtils.getLogger(getClass());
        this.repNode = repNode;
        this.vlsnIndex = repNode.getVLSNIndex();
        this.namedChannel = namedChannel;
        this.protocol = protocol;
        this.repImpl = repNode.getRepImpl();
    }

    public long getMatchedVLSNTime() {
        return matchedVLSNTime;
    }

    public VLSN getMatchedVLSN() {
        return matchpointVLSN;
    }

    /**
     * The replica's side of the protocol.
     * @throws InterruptedException
     * @throws InsufficientLogException
     */
    public void execute(LocalCBVLSNTracker cbvlsnTracker)
        throws IOException,
               DatabaseException,
               InterruptedException,
               InsufficientLogException {

        final long startTime = System.currentTimeMillis();
        String feederName = namedChannel.getNameIdPair().getName();
        LoggerUtils.info(logger, repImpl,
                         "Replica-feeder " + feederName + 
                         " syncup started. Replica range: " +
                         repNode.getVLSNIndex().getRange());

        /* Prohibit global CBVLSN update. */
        repNode.syncupStarted();
        try {

            /*
             * Find a replication stream matchpoint and a place to start
             * the replication stream. If the feeder cannot service this
             * protocol because it has run out of replication stream,
             * findMatchpoint will throw a InsufficientLogException. If the
             */
            VLSNRange range = vlsnIndex.getRange();
            findMatchpoint(range);

            /*
             * If we can't rollback to the found matchpoint, verifyRollback
             * with throw an exception.
             */
            verifyRollback(range);

            replay.rollback(matchpointVLSN, matchpointLsn);

            /* Update the vlsnIndex. */
            VLSN startVLSN = matchpointVLSN.getNext();
            vlsnIndex.truncateFromTail(startVLSN, matchpointLsn);

            protocol.write(protocol.new StartStream(startVLSN), namedChannel);
            LoggerUtils.info(logger, repImpl,
                             "Replica-feeder " + feederName + 
                             " start stream at VLSN: " + startVLSN);

            /*
             * Initialize this node's local CBVLSN while global CBVLSN updates
             * are prohibited.
             */
            cbvlsnTracker.registerMatchpoint(startVLSN);
            if (syncupEndHook != null) {
                syncupEndHook.doHook();
            }
        } finally {
            repNode.syncupEnded();
            LoggerUtils.info
                (logger, repImpl,
                 String.format
                 ("Replica-feeder " + feederName + 
                  " syncup ended. Elapsed time: %,dms",
                  (System.currentTimeMillis() - startTime)));
        }
    }

    /**
     * A matchpoint has been found. What happens next depends on the position
     * of the matchpoint and its relationship to the last transaction end
     * record.
     *
     * In following table,
     *    M = some non-null matchpoint VLSN value,
     *    T = some non-null last txn end value
     *    S = some non-null last sync value
     *
     * txn end                last sync   found        action
     *  VLSN                  VLSN        matchpoint
     * ----------             ---------   ---------    ------------------------
     * NULL_VLSN              NULL_VLSN   NULL_VLSN    rollback everything
     * NULL_VLSN              NULL_VLSN      M         can't occur
     * NULL_VLSN                 S        NULL_VLSN    rollback everything
     * NULL_VLSN                 S           M         rollback to M
     *   T                    NULL_VLSN   NULL_VLSN    can't occur
     *   T                    NULL_VLSN      M         can't occur
     *   T                       S        NULL_VLSN    network restore, though
     *                                                 could also do hard recov
     *   T <= M                  S           M         rollback to matchpoint
     *   T > M, truncate ok      S           M         hard recovery
     *   T > M, truncate not ok  S           M         network restore
     * @throws IOException
     */
    private void verifyRollback(VLSNRange range)
        throws RollbackException, InsufficientLogException, IOException {

        VLSN lastTxnEnd = range.getLastTxnEnd();
        VLSN lastSync = range.getLastSync();

        /*
         * If the lastTxnEnd VLSN is null, we don't have to worry about hard
         * recovery.
         */
        if (lastTxnEnd.equals(NULL_VLSN)) {

            if (range.getLastSync().equals(NULL_VLSN) &&
                !(matchpointVLSN.equals(NULL_VLSN))) {
                throw EnvironmentFailureException.unexpectedState
                    (repNode.getRepImpl(), "Shouldn't be possible to find a "+
                     "matchpoint of " + matchpointVLSN +
                     " when the sync VLSN is "+ " null" + range);
            }

            /* We'll be doing a normal rollback. */
            return;
        }

        if (lastSync.equals(NULL_VLSN)) {
            throw EnvironmentFailureException.unexpectedState
                (repNode.getRepImpl(),
                 "Shouldn't be possible to have a non-null sync VLSN when the "
                 + " lastTxnVLSN is null" + range);
        }

        /*
         * There is a lastTxnEnd VLSN here, so check if the found matchpoint
         * precedes it. If it doesn, we can't rollback.
         */
        if (matchpointVLSN.equals(NULL_VLSN)) {


            /*
             * We could actually also try to do a hard recovery, but
             * for now, let's assume that it would cost less to
             * copy the log files.
             */
            throw setupLogRefresh(matchpointVLSN);
        }

        /*
         * The matchpoint is after or equal to the last txn end, no problem
         * with doing a normal rollback.
         */
        if (lastTxnEnd.compareTo(matchpointVLSN) <= 0) {
            return;
        }

        /*
         * We're rolling back past a commit or abort. If permissible, truncate
         * the log and run a hard recovery. If the matchpoint precedes a
         * checkpoint which deleted log files, the truncation is not
         * permissible, and we have to do a network restore instead.
         */
        if (passedCheckpointEnd) {
            throw setupLogRefresh(matchpointVLSN);
        }

        /* Do truncation and hard recovery */
        throw setupHardRecovery(range);
    }

    /**
     * Find a matchpoint, which is a log entry in the replication stream which
     * is the same on feeder and replica. Assign the matchpointVLSN field. The
     * matchpoint log entry must be be tagged with an environment id. If no
     * matching entry is found, the matchpoint is effectively the NULL_VLSN.
     *
     * To determine the matchpoint, exchange messages with the feeder and
     * compare log entries. If the feeder does not have enough log entries,
     * throw InsufficientLogException.
     * @throws InterruptedException
     * @throws InsufficientLogException
     */
    private void findMatchpoint(VLSNRange range)
        throws IOException,
               InterruptedException,
               InsufficientLogException {

        VLSN candidateMatchpoint = range.getLastSync();
        if (candidateMatchpoint.equals(NULL_VLSN)) {

            /*
             * If the replica has no sync-able log entries at all, the
             * matchpoint is the NULL_VLSN, and we should start the replication
             * stream at VLSN 1. Check if the feeder has the VLSN 1. If it
             * doesn't, getFeederRecord() will throw a
             * InsufficientLogException. We can assume that a non-cleaned
             * feeder always has VLSN 1, because a ReplicatedEnvironment always
             * creates a few replicated vlsns, such as the name db, at
             * initial startup.
             */
            getFeederRecord(range, VLSN.FIRST_VLSN,
                            false /*acceptAlternative*/);
            return;
        }

        /*
         * CandidateMatchpoint is not null, so ask the feeder for the log
         * record at that vlsn.
         */
        InputWireRecord feederRecord =
            getFeederRecord(range, candidateMatchpoint,
                            true /*acceptAlternative*/);

        /*
         * The feeder may have suggested an alternative matchpoint, so reset
         * candidate matchpoint.
         */
        candidateMatchpoint = feederRecord.getVLSN();
        if (logger.isLoggable(Level.FINEST)) {
            LoggerUtils.finest(logger, repImpl,
                               "Matchpoint: " + candidateMatchpoint);
        }
        /* Start comparing feeder records to replica records. */
        ReplicaSyncupReader backwardsReader =
            setupBackwardsReader(candidateMatchpoint);
        OutputWireRecord replicaRecord =
            backwardsReader.scanBackwards(candidateMatchpoint);

        if (!replicaRecord.match(feederRecord)) {

            while (true) {

                replicaRecord = backwardsReader.findPrevSyncEntry();

                if (replicaRecord == null) {

                    /*
                     * The search for the previous sync log entry went past our
                     * available contiguous VLSN range, so there is no
                     * matchpoint.
                     */
                    throw setupLogRefresh(candidateMatchpoint);
                }

                /*
                 * Ask the feeder for the record. If the feeder doesn't have
                 * it, we'll throw out and do a network restore.
                 */
                candidateMatchpoint = replicaRecord.getVLSN();
                if (logger.isLoggable(Level.FINEST)) {
                    LoggerUtils.finest(logger, repImpl, "Next matchpoint: " + 
                                       candidateMatchpoint);
                }
                feederRecord = getFeederRecord(range, candidateMatchpoint,
                                               false);

                if (replicaRecord.match(feederRecord)) {
                    break;
                }
            }
        }

        /* We've found the matchpoint. */
        matchedVLSNTime = replicaRecord.getTimeStamp();
        matchpointVLSN = candidateMatchpoint;
        matchpointLsn = backwardsReader.getLastLsn();
        passedCheckpointEnd = backwardsReader.getPassedCheckpointEnd();
    }

    private ReplicaSyncupReader setupBackwardsReader(VLSN candidateMatchpoint)
        throws IOException {

        EnvironmentImpl envImpl = repNode.getRepImpl();
        int readBufferSize = envImpl.getConfigManager().
            getInt(EnvironmentParams.LOG_ITERATOR_READ_SIZE);
        return new ReplicaSyncupReader
            (envImpl,
             repNode.getVLSNIndex(),
             envImpl.getFileManager().getLastUsedLsn(),
             readBufferSize,
             repNode.getNameIdPair(),
             candidateMatchpoint,
             DbLsn.makeLsn
             (repNode.getCleanerBarrierFile(),  0));
    }

    /**
     * Ask the feeder for information to add to InsufficientLogException,
     * and then throw the exception.
     *
     * The endVLSN marks the last VLSN that this node will want from
     * the network restore. That information helps ensure that the restore
     * source has enough vlsns to satisfy this replica.
     *
     * The replication node list identifies possible log provider members.
     * @throws IOException
     */
    private InsufficientLogException setupLogRefresh(VLSN failedMatchpoint)
        throws IOException {

        protocol.write(protocol.new RestoreRequest(failedMatchpoint),
                       namedChannel);
        RestoreResponse response =
            (RestoreResponse) protocol.read(namedChannel);

        return new InsufficientLogException
            (repNode,
             response.getCBVLSN(),
             new HashSet<ReplicationNode>(Arrays.asList
                                          (response.getLogProviders())));
    }


    /**
     * Hard recovery:  truncate the files, repeat recovery.
     * If this hard recovery came about before the ReplicatedEnvironment was
     * fully instantiated, we will recreate the environment under the
     * covers. If this came while the replica was up and supporting existing
     * Environment handles, we must invalidate the environment, and ask the
     * application to reopen.
     * @throws IOException
     */
    public RollbackException setupHardRecovery(VLSNRange range)

        throws IOException {

        /* Creating the exception invalidates the environment. */
        RollbackException r = new RollbackException(repImpl,
                                                    range.getLastTxnEnd(),
                                                    matchpointVLSN,
                                                    matchpointLsn);

        /* Truncate after the environment is invalidated. */
        repImpl.getFileManager().truncateLog
            (DbLsn.getFileNumber(matchpointLsn),
             DbLsn.getFileOffset(matchpointLsn));

        return r;
    }

    /**
     * Request a log entry from the feeder at this VLSN. The Feeder will only
     * return the log record or say that it isn't available.
     *
     * @throws InsufficientLogException
     */
    private InputWireRecord getFeederRecord(VLSNRange range,
                                            VLSN requestVLSN,
                                            boolean acceptAlternative)
        throws IOException, InsufficientLogException {

        /* Ask the feeder for the matchpoint log record. */
        protocol.write(protocol.new EntryRequest(requestVLSN), namedChannel);

        /*
         * Expect
         *  a) the requested log record
         *  b) message that says this feeder doesn't have RequestVLSN
         *  c) if acceptAlternative == true and the feeder didn't have
         *     requestVLSN, but had an earlier entry, the feeder may send an
         *      earlier, alternative matchpoint
         */
        Message message = protocol.read(namedChannel);
        if (message instanceof Entry) {
            Entry entry = (Entry) message;
            return entry.getWireRecord();
        }

        if (message instanceof EntryNotFound) {
            throw setupLogRefresh(requestVLSN);
        }

        if ((acceptAlternative) &&
            (message instanceof AlternateMatchpoint)) {

            AlternateMatchpoint alt = (AlternateMatchpoint) message;
            InputWireRecord feederRecord = alt.getAlternateWireRecord();
            VLSN altMatchpoint = feederRecord.getVLSN();
            if (range.getFirst().compareTo(altMatchpoint) > 0) {

                /*
                 * The feeder suggest a different matchpoint, but it's outside
                 * the replica's range. Give up and do a network restore.
                 */
                throw setupLogRefresh(altMatchpoint);
            }
            return feederRecord;
        }

        throw EnvironmentFailureException.unexpectedState
            (repNode.getRepImpl(),
             "Sent EntryRequest, got unexpected response of " + message);
    }

    public static TestHook<Object> getSyncupEndHook() {
        return syncupEndHook;
    }

    public static void setSyncupEndHook(TestHook<Object> syncupEndHook) {
        ReplicaFeederSyncup.syncupEndHook = syncupEndHook;
    }

    public interface TestHook<T> {

        public void doHook() throws InterruptedException;

    }

}
